@@ -20,25 +20,25 @@ Methods to annotate the Abstract Syntax Tree (AST) of an IDP-Z3 program.
 
 """
 
-from copy import copy
+from copy import copy, deepcopy
 from itertools import chain
 import string
 from typing import Dict
 
-from .Parse import (Vocabulary, Import, TypeDeclaration, Declaration, Type,
-                    SymbolDeclaration, Symbol, VarDeclaration,
-                    TheoryBlock, Definition, Rule,
-                    Structure, SymbolInterpretation, Enumeration, FunctionEnum,
-                    Tuple, ConstructedFrom, Display)
-from .Expression import (Expression, Constructor, AIfExpr, AQuantification, Quantee,
-                         ARImplication, AImplication, AEquivalence,
-                         Operator, AComparison, AUnary, AAggregate,
-                         AppliedSymbol, UnappliedSymbol, Variable, Brackets,
-                         FALSE, SymbolExpr, Number, NOT, EQUALS, AND, OR, TRUE, FALSE,
-                         IMPLIES, RIMPLIES, EQUIV, FORALL, EXISTS, Extension)
+from .Parse import (Vocabulary, Import, TypeDeclaration, Declaration,
+    SymbolDeclaration, VarDeclaration, TheoryBlock, Definition, Rule,
+    Structure, SymbolInterpretation, Enumeration, FunctionEnum,
+    TupleIDP, ConstructedFrom, Display)
+from .Expression import (Expression, Symbol, SYMBOL, Type, TYPE,
+    Constructor, CONSTRUCTOR, AIfExpr, IF, AQuantification, Quantee,
+    ARImplication, AImplication, AEquivalence,
+    Operator, AComparison, AUnary, AAggregate,
+    AppliedSymbol, UnappliedSymbol, Variable, VARIABLE, Brackets,
+    FALSE, SymbolExpr, Number, NOT, EQUALS, AND, OR, FALSE,
+    IMPLIES, RIMPLIES, EQUIV, FORALL, EXISTS, Extension)
 
 from .utils import (BOOL, INT, REAL, DATE, CONCEPT, RESERVED_SYMBOLS,
-                    OrderedSet, IDPZ3Error, DEF_SEMANTICS, Semantics)
+    OrderedSet, IDPZ3Error, DEF_SEMANTICS, Semantics)
 
 
 # Class Vocabulary  #######################################################
@@ -61,9 +61,9 @@ def annotate(self, idp):
             s.check(s.name not in temp or s.name in RESERVED_SYMBOLS,
                     f"Duplicate declaration of {s.name}")
             temp[s.name] = s
-    temp[CONCEPT].constructors=([Constructor(name=f"`{s}")
+    temp[CONCEPT].constructors=([CONSTRUCTOR(f"`{s}")
                                  for s in [BOOL, INT, REAL, DATE, CONCEPT]]
-                               +[Constructor(name=f"`{s.name}")
+                               +[CONSTRUCTOR(f"`{s.name}")
                                  for s in temp.values()
                                  if s.name not in RESERVED_SYMBOLS
                                  and type(s) in Declaration.__args__])
@@ -75,7 +75,7 @@ def annotate(self, idp):
 
     concepts = self.symbol_decls[CONCEPT]
     for constructor in concepts.constructors:
-        constructor.symbol = (Symbol(name=constructor.name[1:])
+        constructor.symbol = (SYMBOL(constructor.name[1:])
                                 .annotate(self, {}))
 
     # populate .map of CONCEPT
@@ -88,6 +88,7 @@ Vocabulary.annotate = annotate
 # Class TypeDeclaration  #######################################################
 
 def annotate(self, voc):
+    self.voc = voc
     self.check(self.name not in voc.symbol_decls,
                 f"duplicate declaration in vocabulary: {self.name}")
     voc.symbol_decls[self.name] = self
@@ -190,12 +191,12 @@ def annotate(self, voc, q_vars):
     for r in self.rules:
         decl = voc.symbol_decls[r.definiendum.decl.name]
         if decl.name not in self.def_vars:
-            name = f"${decl.name}$"
-            q_v = {f"${decl.name}!{str(i)}$":
-                    Variable(name=f"${decl.name}!{str(i)}$", sort=sort)
+            name = f"{decl.name}_"
+            q_v = {f"{decl.name}{str(i)}_":
+                    VARIABLE(f"{decl.name}{str(i)}_", sort)
                     for i, sort in enumerate(decl.sorts)}
             if decl.out.name != BOOL:
-                q_v[name] = Variable(name=name, sort=decl.out)
+                q_v[name] = VARIABLE(name, decl.out)
             self.def_vars[decl.name] = q_v
         new_rule = r.rename_args(self.def_vars[decl.name])
         self.canonicals.setdefault(decl, []).append(new_rule)
@@ -258,7 +259,7 @@ def get_instantiables(self,
                     new = r.body.split_equivalences()
                     bodies.append(new)
                     if for_explain:
-                        new = new.copy().add_level_mapping(rule.parent.level_symbols,
+                        new = deepcopy(new).add_level_mapping(rule.parent.level_symbols,
                                              rule.definiendum, False, False)
                         out.append(RIMPLIES([head, new], r.annotations))
 
@@ -270,10 +271,10 @@ def get_instantiables(self,
                     out = [EQUIV([head, all_bodies], self.annotations)]
             else:
                 if not out:  # no reverse implication yet
-                    new = all_bodies.copy().add_level_mapping(rule.parent.level_symbols,
+                    new = deepcopy(all_bodies).add_level_mapping(rule.parent.level_symbols,
                                              rule.definiendum, False, False)
-                    out = [RIMPLIES([head.copy(), new], self.annotations)]
-                all_bodies = all_bodies.copy().add_level_mapping(rule.parent.level_symbols,
+                    out = [RIMPLIES([deepcopy(head), new], self.annotations)]
+                all_bodies = deepcopy(all_bodies).add_level_mapping(rule.parent.level_symbols,
                                         rule.definiendum, True, True)
                 out.append(IMPLIES([head, all_bodies], self.annotations))
             result[decl] = out
@@ -284,6 +285,7 @@ Definition.get_instantiables = get_instantiables
 # Class Rule  #######################################################
 
 def annotate(self, voc, q_vars):
+    self.original = copy(self)
     self.check(not self.definiendum.symbol.is_intentional(),
                 f"No support for intentional objects in the head of a rule: "
                 f"{self}")
@@ -366,7 +368,7 @@ def annotate(self, block):
     """
     voc = block.voc
     self.block = block
-    self.symbol = Symbol(name=self.name).annotate(voc, {})
+    self.symbol = SYMBOL(self.name).annotate(voc, {})
     enumeration = self.enumeration  # shorthand
 
     # create constructors if it is a type enumeration
@@ -411,13 +413,13 @@ def annotate(self, voc):
 Enumeration.annotate = annotate
 
 
-# Class Tuple  #######################################################
+# Class TupleIDP  #######################################################
 
 def annotate(self, voc):
     self.args = [arg.annotate(voc, {}) for arg in self.args]
     self.check(all(a.value is not None for a in self.args),
                 f"Tuple must be ground : ({self})")
-Tuple.annotate = annotate
+TupleIDP.annotate = annotate
 
 
 # Class ConstructedFrom  #######################################################
@@ -426,7 +428,7 @@ def annotate(self, voc):
     for c in self.constructors:
         for i, ts in enumerate(c.sorts):
             if ts.accessor is None:
-                ts.accessor = Symbol(name=f"{c.name}_{i}")
+                ts.accessor = SYMBOL(f"{c.name}_{i}")
             if ts.accessor.name in self.accessors:
                 self.check(self.accessors[ts.accessor.name] == i,
                            "Accessors used at incompatible indices")
@@ -443,13 +445,13 @@ def annotate(self, voc):
         self.check(a.type in voc.symbol_decls,
                    f"Unknown type: {a.type}" )
         a.decl = SymbolDeclaration(annotations='', name=a.accessor,
-                                   sorts=[Type(name=self.type)],
-                                   out=Type(name=a.type))
+                                   sorts=[TYPE(self.type)],
+                                   out=TYPE(a.type))
         a.decl.annotate(voc)
     self.tester = SymbolDeclaration(annotations='',
-                                    name=Symbol(name=f"is_{self.name}"),
-                                    sorts=[Type(name=self.type)],
-                                    out=Type(name=BOOL))
+                                    name=SYMBOL(f"is_{self.name}"),
+                                    sorts=[TYPE(self.type)],
+                                    out=TYPE(BOOL))
     self.tester.annotate(voc)
 Constructor.annotate = annotate
 
@@ -462,8 +464,8 @@ def annotate(self, idp):
     # add display predicates
 
     viewType = TypeDeclaration(name='_ViewType',
-        constructors=[Constructor(name='normal'),
-                        Constructor(name='expanded')])
+        constructors=[CONSTRUCTOR('normal'),
+                        CONSTRUCTOR('expanded')])
     viewType.annotate(self.voc)
 
     # Check the AST for any constructors that belong to open types.
@@ -484,23 +486,23 @@ def annotate(self, idp):
         open_type = TypeDeclaration(name=type_name,
                                     constructors=constructors)
         open_type.annotate(self.voc)
-        open_types[name] = Symbol(name=type_name)
+        open_types[name] = SYMBOL(type_name)
 
     for name, out in [
-        ('expand', Symbol(name=BOOL)),
-        ('hide', Symbol(name=BOOL)),
-        ('view', Symbol(name='_ViewType')),
-        ('moveSymbols', Symbol(name=BOOL)),
-        ('optionalPropagation', Symbol(name=BOOL)),
-        ('manualPropagation', Symbol(name=BOOL)),
-        ('optionalRelevance', Symbol(name=BOOL)),
-        ('manualRelevance', Symbol(name=BOOL)),
+        ('expand', SYMBOL(BOOL)),
+        ('hide', SYMBOL(BOOL)),
+        ('view', SYMBOL('_ViewType')),
+        ('moveSymbols', SYMBOL(BOOL)),
+        ('optionalPropagation', SYMBOL(BOOL)),
+        ('manualPropagation', SYMBOL(BOOL)),
+        ('optionalRelevance', SYMBOL(BOOL)),
+        ('manualRelevance', SYMBOL(BOOL)),
         ('unit', open_types['unit']),
         ('heading', open_types['heading']),
-        ('noOptimization', Symbol(name=BOOL))
+        ('noOptimization', SYMBOL(BOOL))
     ]:
         symbol_decl = SymbolDeclaration(annotations='',
-                                        name=Symbol(name=name),
+                                        name=SYMBOL(name),
                                         sorts=[out] if out else [], out=out)
         symbol_decl.annotate(self.voc)
 
@@ -549,6 +551,8 @@ Expression.annotate1 = annotate1
 
 def annotate1(self):
     self.type = self.sub_exprs[AIfExpr.THEN].type
+    if not self.type:
+        self.type = self.sub_exprs[AIfExpr.ELSE].type
     return Expression.annotate1(self)
 AIfExpr.annotate1 = annotate1
 
@@ -561,12 +565,15 @@ def annotate(self, voc, q_vars):
     for q in self.quantees:
         q.annotate(voc, q_vars)
         for vars in q.vars:
-            for var in vars:
+            for i, var in enumerate(vars):
                 self.check(var.name not in voc.symbol_decls
                            or type(voc.symbol_decls[var.name]) == VarDeclaration,
                     f"the quantified variable '{var.name}' cannot have"
                     f" the same name as another symbol")
-                var.sort = q.sub_exprs[0] if q.sub_exprs else None
+                if len(vars) == 1:
+                    var.sort = q.sub_exprs[0] if q.sub_exprs else None
+                else:  # VarTuple
+                    var.sort = q.sub_exprs[0].decl.sorts[i]
                 var_decl = voc.symbol_decls.get(var.name.rstrip(string.digits), None)
                 if var_decl:
                     subtype = var_decl.subtype
@@ -576,6 +583,8 @@ def annotate(self, voc, q_vars):
                         f"{var.sort.name if var.sort else ''}")
                     if var.sort is None:
                         q.sub_exprs = [subtype.annotate(voc, {})]
+                        var.sort = q.sub_exprs[0]
+                var.type = var.sort.decl.name if var.sort and var.sort.decl else ''
                 q_v[var.name] = var
     self.sub_exprs = [e.annotate(voc, q_v) for e in self.sub_exprs]
     return self.annotate1()
@@ -628,9 +637,9 @@ AEquivalence.annotate1 = annotate1
 
 def annotate(self, voc, q_vars):
     # reverse the implication
-    self.sub_exprs.reverse()
-    out = AImplication(sub_exprs=self.sub_exprs,
+    out = AImplication(None, sub_exprs=list(reversed(list(self.sub_exprs))),
                         operator=['⇒']*len(self.operator))
+    out.original = self
     if hasattr(self, "block"):
         out.block = self.block
     return out.annotate(voc, q_vars)
@@ -667,9 +676,8 @@ def annotate(self, voc, q_vars):
     if not self.annotated:
         assert len(self.sub_exprs) == 1, "Internal error"
         if self.aggtype == "#":
-            self.sub_exprs = [AIfExpr.make(self.sub_exprs[0],
-                                          Number(number='1'),
-                                          Number(number='0'))]
+            self.sub_exprs = [IF(self.sub_exprs[0], Number(number='1'),
+                                 Number(number='0'))]
             self.type = INT
         else:
             self.type = self.sub_exprs[0].type
@@ -678,6 +686,7 @@ def annotate(self, voc, q_vars):
                 # is replaced by `_*(y)` with the following co-constraint:
                 #     !y in T: ( ?x in type: term(x) = _*(y)
                 #                !x in type: term(x) =< _*(y).
+                self.check(self.type, f"Can't infer type of {self}")
                 name = "_" + self.str
                 if name in voc.symbol_decls:
                     symbol_decl = voc.symbol_decls[name]
@@ -686,20 +695,20 @@ def annotate(self, voc, q_vars):
                     symbol_decl = SymbolDeclaration.make(
                         "_"+self.str, # name `_ *`
                         len(q_vars),  # arity
-                        [Type(name=v.sort.code) for v in q_vars.values()],
-                        Type(name=self.type)).annotate(voc)    # output_domain
+                        [TYPE(v.sort.code) for v in q_vars.values()],
+                        TYPE(self.type)).annotate(voc)    # output_domain
                     to_create = True
-                symbol = Symbol(name=symbol_decl.name)
+                symbol = SYMBOL(symbol_decl.name)
                 applied = AppliedSymbol.make(symbol, q_vars.values())
                 applied = applied.annotate(voc, q_vars)
 
                 if to_create:
                     coc1 = EXISTS(self.quantees,
-                                EQUALS([applied.copy(), self.sub_exprs[0]]))
+                                EQUALS([deepcopy(applied), self.sub_exprs[0]]))
                     op = '≤' if self.aggtype == "min" else '≥'
-                    coc2 = FORALL(self.quantees.copy(),
+                    coc2 = FORALL(deepcopy(self.quantees),
                                 AComparison.make(op,
-                                        [applied.copy(), self.sub_exprs[0].copy()]))
+                                    deepcopy([applied, self.sub_exprs[0]])))
                     coc = AND([coc1, coc2])
                     quantees = [Quantee.make(v, v.sort) for v in q_vars.values()]
                     applied.co_constraint = FORALL(quantees, coc).annotate(voc, q_vars)
