@@ -30,7 +30,7 @@ from z3 import (Z3Exception, Datatype, DatatypeRef, ExprRef,
                 Or, Not, And, ForAll, Exists, Sum, If,
                 BoolVal, RatVal, IntVal)
 
-from idp_engine.Parse import TypeDeclaration, SymbolDeclaration, Tuple
+from idp_engine.Parse import TypeDeclaration, SymbolDeclaration, TupleIDP, Ranges, IntRange, RealRange, DateRange
 from idp_engine.Expression import (Constructor, Expression, AIfExpr,
                                    AQuantification, Operator, Symbol,
                                    ADisjunction, AConjunction, AComparison,
@@ -73,9 +73,11 @@ def translate(self, problem: "Theory"):
                 elif c.range:
                     for e in c.range:
                         self.map[str(e)] = e
-        else: # list of numbers
+        elif type(self.interpretation.enumeration) in [Ranges, IntRange, RealRange, DateRange]: # list of numbers
             out = (IntSort(problem.ctx) if self.interpretation.enumeration.type in [INT, DATE] else
                    RealSort(problem.ctx))
+        else:  # empty type --> don't care
+            out = IntSort(problem.ctx)
         problem.z3[self.name] = out
     return out
 TypeDeclaration.translate = translate
@@ -97,11 +99,11 @@ def translate(self, problem: "Theory"):
 SymbolDeclaration.translate = translate
 
 
-# class Tuple  ###########################################################
+# class TupleIDP  ###########################################################
 
 def translate(self, problem: "Theory"):
     return [arg.translate(problem) for arg in self.args]
-Tuple.translate = translate
+TupleIDP.translate = translate
 
 # class Constructor  ###########################################################
 
@@ -193,7 +195,7 @@ def translate1(self, problem: "Theory", vars={}):
         for q in self.quantees:
             for vars in q.vars:
                 for v in vars:
-                    translated = FreshConst(v.sort.decl.translate(problem))
+                    translated = FreshConst(v.sort.decl.base_type.translate(problem))
                     finalvars[v.str] = translated
         forms = [f.translate(problem, finalvars) for f in self.sub_exprs]
 
@@ -210,23 +212,23 @@ AQuantification.translate1 = translate1
 # Class Operator  #######################################################
 
 Operator.MAP = {'∧': lambda x, y: And(x, y),
-                      '∨': lambda x, y: Or(x, y),
-                      '⇒': lambda x, y: Or(Not(x), y),
-                      '⇐': lambda x, y: Or(x, Not(y)),
-                      '⇔': lambda x, y: x == y,
-                      '+': lambda x, y: x + y,
-                      '-': lambda x, y: x - y,
-                      '⨯': lambda x, y: x * y,
-                      '/': lambda x, y: x / y,
-                      '%': lambda x, y: x % y,
-                      '^': lambda x, y: x ** y,
-                      '=': lambda x, y: x == y,
-                      '<': lambda x, y: x < y,
-                      '>': lambda x, y: x > y,
-                      '≤': lambda x, y: x <= y,
-                      '≥': lambda x, y: x >= y,
-                      '≠': lambda x, y: x != y
-                      }
+                '∨': lambda x, y: Or(x, y),
+                '⇒': lambda x, y: Or(Not(x), y),
+                '⇐': lambda x, y: Or(x, Not(y)),
+                '⇔': lambda x, y: x == y,
+                '+': lambda x, y: x + y,
+                '-': lambda x, y: x - y,
+                '⨯': lambda x, y: x * y,
+                '/': lambda x, y: x / y,
+                '%': lambda x, y: x % y,
+                '^': lambda x, y: x ** y,
+                '=': lambda x, y: x == y,
+                '<': lambda x, y: x < y,
+                '>': lambda x, y: x > y,
+                '≤': lambda x, y: x <= y,
+                '≥': lambda x, y: x >= y,
+                '≠': lambda x, y: x != y
+                }
 
 
 def translate1(self, problem: "Theory", vars={}):
