@@ -35,19 +35,18 @@ This module monkey-patches the ASTNode class and sub-classes.
 
 """
 
-import copy
+from copy import copy, deepcopy
 from itertools import product
-from typing import Dict, Tuple, List, Callable
+from typing import Dict, List, Callable
 
 from .Assignments import Status as S
-from .Parse import (Import, TypeDeclaration,
-                    SymbolDeclaration, Symbol, SymbolInterpretation,
-                    FunctionEnum, Enumeration, Tuple, ConstructedFrom,
-                    Definition, Ranges, ConstructedFrom)
-from .Expression import (AIfExpr, SymbolExpr, Expression, Constructor,
-                    AQuantification, Type, FORALL, IMPLIES, AND, AAggregate,
-                    NOT, AppliedSymbol, UnappliedSymbol, Quantee,
-                    Variable, TRUE, FALSE, Number, Extension)
+from .Parse import (Import, TypeDeclaration, SymbolDeclaration,
+    SymbolInterpretation, FunctionEnum, Enumeration, TupleIDP, ConstructedFrom,
+    Definition, ConstructedFrom)
+from .Expression import (Symbol, SYMBOL, AIfExpr, IF, SymbolExpr, Expression, Constructor,
+    AQuantification, Type, FORALL, IMPLIES, AND, AAggregate,
+    NOT, AppliedSymbol, UnappliedSymbol, Quantee,
+    Variable, VARIABLE, TRUE, FALSE, Number, Extension)
 from .Theory import Theory
 from .utils import (BOOL, RESERVED_SYMBOLS, CONCEPT, OrderedSet, DEFAULT,
                     GOAL_SYMBOL, EXPAND)
@@ -87,7 +86,7 @@ TypeDeclaration.interpret = interpret
 def interpret(self, problem):
     assert all(isinstance(s, Type) for s in self.sorts), 'internal error'
 
-    symbol = Symbol(name=self.name)
+    symbol = SYMBOL(self.name)
     symbol.decl = self
     symbol.type = symbol.decl.type
 
@@ -101,10 +100,10 @@ def interpret(self, problem):
 
     filters = [e[1] for e in extensions]
     def filter(args):
-        out = AND([f([t.copy()]) if f is not None else TRUE
+        out = AND([f([deepcopy(t)]) if f is not None else TRUE
                     for f, t in zip(filters, args)])
         if self.out.decl.name == BOOL:
-            out = AND([out, AppliedSymbol.make(symbol, args).copy()])
+            out = AND([out, deepcopy(AppliedSymbol.make(symbol, args))])
         return out
 
     if self.out.decl.name == BOOL:
@@ -134,10 +133,10 @@ def interpret(self, problem):
         for expr in self.instances.values():
             # add type constraints to problem.constraints
             # ! (x,y) in domain: range(f(x,y))
-            range_condition = self.out.has_element(expr.copy(),
+            range_condition = self.out.has_element(deepcopy(expr),
                                 problem.interpretations, problem.extensions)
             range_condition = range_condition.interpret(problem)
-            constraint = IMPLIES([filter(args), range_condition])
+            constraint = IMPLIES([filter(expr.sub_exprs), range_condition])
             constraint.block = self.block
             constraint.is_type_constraint_for = self.name
             constraint.annotations['reading'] = f"Possible values for {expr}"
@@ -262,11 +261,11 @@ def interpret(self, problem):
                         and type(self.enumeration) == FunctionEnum
                         and self.default.type != BOOL):
                         problem.assignments.assert__(e.formula(), TRUE, status)
-        elif self.sign == ':=':
+        elif self.sign == '≜':
             # add condition that the interpretation is total over the domain
             # ! x in dom(f): enum.contains(x)
             q_vars = { f"${sort.decl.name}!{str(i)}$":
-                       Variable(name=f"${sort.decl.name}!{str(i)}$", sort=sort)
+                       VARIABLE(f"${sort.decl.name}!{str(i)}$", sort)
                        for i, sort in enumerate(decl.sorts)}
             quantees = [Quantee.make(v, v.sort) for v in q_vars.values()]
             expr = self.enumeration.contains(list(q_vars.values()), True)
@@ -292,7 +291,7 @@ def interpret(self, problem):
         if c.range is None:
             self.tuples = None
             return self
-        self.tuples.extend([Tuple(args=[e]) for e in c.range])
+        self.tuples.extend([TupleIDP(args=[e]) for e in c.range])
     return self
 ConstructedFrom.interpret = interpret
 
@@ -373,9 +372,9 @@ def instantiate(self, e0, e1, problem=None):
            f"Internal error: instantiate {e0}"
     if self.value:
         return self
-    out = copy.copy(self)  # shallow copy !
-    out.annotations = copy.copy(out.annotations)
-    out.variables = copy.copy(out.variables)
+    out = copy(self)  # shallow copy !
+    out.annotations = copy(out.annotations)
+    out.variables = copy(out.variables)
     return out.instantiate1(e0, e1, problem)
 Expression.instantiate = instantiate
 
@@ -472,12 +471,10 @@ def _add_filter(q: str, expr: Expression, filter: Callable, args: List[Expressio
         else:  # aggregate
             if isinstance(expr, AIfExpr):  # cardinality
                 # if a then b else 0 -> if (applied & a) then b else 0
-                arg1 = AND([applied,
-                                    expr.sub_exprs[0]])
-                out = AIfExpr.make(arg1, expr.sub_exprs[1],
-                                    expr.sub_exprs[2])
+                arg1 = AND([applied, expr.sub_exprs[0]])
+                out = IF(arg1, expr.sub_exprs[1], expr.sub_exprs[2])
             else:  # sum
-                out = AIfExpr.make(applied, expr, Number(number="0"))
+                out = IF(applied, expr, Number(number="0"))
         return out
     return expr
 
@@ -504,6 +501,7 @@ def interpret(self, problem):
             var = q.vars[0][0]
             self.check(var.name in inferred,
                         f"can't infer type of {var.name}")
+            var.sort = inferred[var.name]
             q.sub_exprs = [inferred[var.name]]
 
     forms = self.sub_exprs
@@ -589,7 +587,7 @@ def interpret(self, problem):
                 simpler.annotations = self.annotations
         elif self.in_enumeration:
             # re-create original Applied Symbol
-            core = AppliedSymbol.make(self.symbol, sub_exprs).copy()
+            core = deepcopy(AppliedSymbol.make(self.symbol, sub_exprs))
             simpler = self.in_enumeration.contains([core], False,
                         interpretations=problem.interpretations, extensions=problem.extensions)
             if 'not' in self.is_enumeration:
