@@ -20,13 +20,8 @@
 Classes to parse an IDP-Z3 theory.
 
 """
-__all__ = ["IDP", "Vocabulary", "Annotations", "Import",
-           "TypeDeclaration",
-           "SymbolDeclaration", "Symbol", "TheoryBlock", "Definition",
-           "Rule", "Structure", "Enumeration", "Tuple",
-           "Display", "Procedure", ]
 
-from copy import copy
+from copy import copy, deepcopy
 from datetime import date
 from enum import Enum
 from itertools import groupby
@@ -38,17 +33,17 @@ from typing import Dict, List, Union, Optional
 
 
 from .Assignments import Assignments
-from .Expression import (Annotations, ASTNode, Constructor, Accessor, Symbol, SymbolExpr,
-                         Expression, AIfExpr, AQuantification, Type, Quantee,
-                         ARImplication, AEquivalence,
-                         AImplication, ADisjunction, AConjunction,
-                         AComparison, ASumMinus, AMultDiv, APower, AUnary,
-                         AAggregate, AppliedSymbol, UnappliedSymbol,
-                         Number, Brackets, Date, Extension,
-                         Variable, TRUEC, FALSEC, TRUE, FALSE, EQUALS, AND, OR, EQUIV)
+from .Expression import (Annotations, ASTNode, Constructor, CONSTRUCTOR, Accessor,
+    Symbol, SYMBOL, SymbolExpr, Expression, AIfExpr, IF, AQuantification,
+    Type, TYPE, Quantee, ARImplication, AEquivalence,
+    AImplication, ADisjunction, AConjunction,
+    AComparison, ASumMinus, AMultDiv, APower, AUnary,
+    AAggregate, AppliedSymbol, UnappliedSymbol,
+    Number, Brackets, Date, Extension,
+    Variable, TRUEC, FALSEC, TRUE, FALSE, EQUALS, AND, OR, EQUIV)
 from .utils import (RESERVED_SYMBOLS, OrderedSet, NEWL, BOOL, INT, REAL, DATE, CONCEPT,
-                    GOAL_SYMBOL, EXPAND, RELEVANT, ABS, IDPZ3Error,
-                    CO_CONSTR_RECURSION_DEPTH, MAX_QUANTIFIER_EXPANSION)
+    GOAL_SYMBOL, EXPAND, RELEVANT, ABS, IDPZ3Error,
+    CO_CONSTR_RECURSION_DEPTH, MAX_QUANTIFIER_EXPANSION)
 
 
 def str_to_IDP(atom, val_string):
@@ -264,23 +259,23 @@ class Vocabulary(ASTNode):
             TypeDeclaration(
                 name=CONCEPT,
                 constructors=[]),
-            SymbolDeclaration(annotations='', name=Symbol(name=GOAL_SYMBOL),
-                              sorts=[Type(name=CONCEPT, ins=[],
-                                             out=Type(name=BOOL))],
-                              out=Type(name=BOOL)),
-            SymbolDeclaration(annotations='', name=Symbol(name=RELEVANT),
-                              sorts=[Type(name=CONCEPT, ins=[],
-                                             out=Type(name=BOOL))],
-                              out=Type(name=BOOL)),
-            SymbolDeclaration(annotations='', name=Symbol(name=ABS),
-                                sorts=[Type(name=INT)],
-                                out=Type(name=INT)),
+            SymbolDeclaration(annotations='', name=SYMBOL(GOAL_SYMBOL),
+                              sorts=[TYPE(CONCEPT, ins=[],
+                                             out=TYPE(BOOL))],
+                              out=TYPE(BOOL)),
+            SymbolDeclaration(annotations='', name=SYMBOL(RELEVANT),
+                              sorts=[TYPE(CONCEPT, ins=[],
+                                             out=TYPE(BOOL))],
+                              out=TYPE(BOOL)),
+            SymbolDeclaration(annotations='', name=SYMBOL(ABS),
+                                sorts=[TYPE(INT)],
+                                out=TYPE(INT)),
             ] + self.declarations
 
     def __str__(self):
         return (f"vocabulary {{{NEWL}"
-                f"{NEWL.join(str(i) for i in self.declarations)}"
-                f"{NEWL}}}{NEWL}")
+                f"    {f'{NEWL}    '.join(str(i) for i in self.declarations)}"
+                f"{NEWL}}}{NEWL}").replace("    \n", "")
 
     def add_voc_to_block(self, block):
         """adds the enumerations in a vocabulary to a theory or structure block
@@ -341,8 +336,8 @@ class TypeDeclaration(ASTNode):
                             kwargs.pop('enumeration'))
 
         self.arity = 1
-        self.sorts = [Symbol(name=self.name)]
-        self.out = Symbol(name=BOOL)
+        self.sorts = [SYMBOL(self.name)]
+        self.out = SYMBOL(BOOL)
         self.type = (self.name if type(enumeration) != Ranges else
                      enumeration.type)  # INT or REAL or DATE
         self.base_type = self
@@ -350,15 +345,16 @@ class TypeDeclaration(ASTNode):
         self.map = {}  # {String: constructor}
 
         self.interpretation = (None if enumeration is None else
-            SymbolInterpretation(name=Symbol(name=self.name), sign=':=',
+            SymbolInterpretation(name=SYMBOL(self.name), sign='≜',
                                  enumeration=enumeration, default=None))
 
     def __str__(self):
         if self.name in RESERVED_SYMBOLS:
             return ''
-        enumeration = (f"{','.join(map(str, self.constructors))}" if self.constructors else
-                       f"{self.enumeration}")
-        return (f"type {self.name} := {{{enumeration}}}")
+        enumeration = (f"{{{','.join(map(str, self.constructors))}}}" if self.constructors else
+                       f"{self.enumeration}" if self.enumeration else
+                       "")
+        return (f"type {self.name} {'' if not enumeration else ':= ' + enumeration}")
 
     def contains_element(self, term: Expression,
                      interpretations: Dict[str, "SymbolInterpretation"],
@@ -445,7 +441,7 @@ class SymbolDeclaration(ASTNode):
         self.sorts = kwargs.pop('sorts')
         self.out = kwargs.pop('out')
         if self.out is None:
-            self.out = Symbol(name=BOOL)
+            self.out = SYMBOL(BOOL)
 
         self.arity = len(self.sorts)
         self.annotations = self.annotations.annotations if self.annotations else {}
@@ -472,7 +468,7 @@ class SymbolDeclaration(ASTNode):
         args = '⨯'.join(map(str, self.sorts)) if 0 < len(self.sorts) else ''
         return (f"{self.name}: "
                 f"{ '('+args+')' if args else '()'}"
-                f" -> {self.out.name}")
+                f" → {self.out.name}")
 
     def __repr__(self):
         return str(self)
@@ -534,6 +530,9 @@ class VarDeclaration(ASTNode):
         self.name = kwargs.pop('name').name
         self.subtype = kwargs.pop('subtype')
 
+    def __str__(self):
+        return f"var {self.name} ∈ {self.subtype}"
+
 Declaration = Union[TypeDeclaration, SymbolDeclaration]
 
 
@@ -677,12 +676,12 @@ class Definition(ASTNode):
             if key not in symbs or key in self.level_symbols:
                 continue
 
-            real = Type(name=REAL)
+            real = TYPE(REAL)
             real.decl = voc.symbol_decls[REAL]
             symbdec = SymbolDeclaration.make(
                 "_"+str(self.id)+"lvl_"+key.name,
                 key.arity, key.sorts, real)
-            self.level_symbols[key] = Symbol(name=symbdec.name)
+            self.level_symbols[key] = SYMBOL(symbdec.name)
             self.level_symbols[key].decl = symbdec
 
         for decl in self.level_symbols.keys():
@@ -717,9 +716,10 @@ class Rule(ASTNode):
             self.body = TRUE
 
     def __repr__(self):
-        return (f"Rule:∀{','.join(str(q) for q in self.quantees)}: "
+        return (f"∀ {','.join(str(q) for q in self.quantees)}: "
                 f"{self.definiendum} "
-                f"⇔{str(self.body)}")
+                f"{(' = ' + str(self.out)) if self.out else ''}"
+                f"← {str(self.body)}")
 
     def instantiate_definition(self, new_args, theory):
         """Create an instance of the definition for new_args, and interpret it for theory.
@@ -733,7 +733,7 @@ class Rule(ASTNode):
         """
 
         #TODO assert self.is_whole_domain == False
-        out = self.body.copy()  # in case there are no arguments
+        out = deepcopy(self.body)  # in case there are no arguments
         instance = AppliedSymbol.make(self.definiendum.symbol, new_args)
         instance.in_head = True
         if self.definiendum.decl.type == BOOL:  # a predicate
@@ -804,10 +804,11 @@ class SymbolInterpretation(ASTNode):
         if not self.enumeration:
             self.enumeration = Enumeration(tuples=[])
 
-        self.sign = ':⊇' if self.sign == ':>=' else self.sign
-        self.check(self.sign == ':=' or
+        self.sign = ('⊇' if self.sign == '>>' else
+                     '≜' if self.sign == ':=' else self.sign)
+        self.check(self.sign == '≜' or
                    (type(self.enumeration) == FunctionEnum and self.default is None),
-                   "':⊇' can only be used with a functional enumeration ('→') without else clause")
+                   "'⊇' can only be used with a functional enumeration ('→') without else clause")
 
         self.symbol = None
         self.is_type_enumeration = None
@@ -818,7 +819,7 @@ class SymbolInterpretation(ASTNode):
 
         This is a recursive function.
 
-        Example: assume `f:>={(1,2)->A, (1, 3)->B, (2,1)->C}` and `args=[g(1),2)].
+        Example: assume `f>>{(1,2)->A, (1, 3)->B, (2,1)->C}` and `args=[g(1),2)].
         The returned expression is:
         ```
         if g(1) = 1 then A
@@ -834,7 +835,7 @@ class SymbolInterpretation(ASTNode):
 
             args (List(Expression)): interpreted arguments applied to the symbol (ex: `g(1),2`)
 
-            tuples (OrderedSet[Tuple], optional): relevant tuples for this iteration.
+            tuples (OrderedSet[TupleIDP], optional): relevant tuples for this iteration.
                 Initialized with `[[1,2,A], [1,3,B], [2,1,C]]`
 
         Returns:
@@ -870,7 +871,7 @@ class SymbolInterpretation(ASTNode):
             else:
                 for val, tuples2 in groups:
                     tuples = list(tuples2)
-                    out = AIfExpr.make(
+                    out = IF(
                         EQUALS([args[rank], tuples[0].args[rank]]),
                         self.interpret_application(rank+1,
                                                    applied, args, tuples),
@@ -884,7 +885,7 @@ class Enumeration(ASTNode):
     Used for predicates, or types without n-ary constructors.
 
     Attributes:
-        tuples (OrderedSet[Tuple]): OrderedSet of Tuple of Expression
+        tuples (OrderedSet[TupleIDP]): OrderedSet of TupleIDP of Expression
 
         sorted_tuples: a sorted list of tuples
 
@@ -900,13 +901,14 @@ class Enumeration(ASTNode):
         self.lookup = {}
         if all(len(c.args) == 1 and type(c.args[0]) == UnappliedSymbol
                for c in self.tuples):
-            self.constructors = [Constructor(name=c.args[0].name)
+            self.constructors = [CONSTRUCTOR(c.args[0].name)
                                  for c in self.tuples]
         else:
             self.constructors = None
 
     def __repr__(self):
-        return ", ".join([repr(t) for t in self.tuples])
+        return (f'{{{", ".join([repr(t) for t in self.tuples])}}}' if self.tuples else
+                f'{{{", ".join([repr(t) for t in self.constructors])}}}')
 
     def contains(self, args, function, arity=None, rank=0, tuples=None,
                  interpretations: Dict[str, "SymbolInterpretation"]=None,
@@ -939,8 +941,7 @@ class Enumeration(ASTNode):
             out = FALSE
             for val, tuples2 in groups:
                 tuples = list(tuples2)
-                out = AIfExpr.make(
-                    EQUALS([args[rank], tuples[0].args[rank]]),
+                out = IF(EQUALS([args[rank], tuples[0].args[rank]]),
                     self.contains(args, function, arity, rank+1, tuples,
                             interpretations=interpretations, extensions=extensions),
                     out)
@@ -980,7 +981,7 @@ class ConstructedFrom(Enumeration):
     """Represents a 'constructed from' enumeration of constructors
 
     Attributes:
-        tuples (OrderedSet[Tuple]): OrderedSet of tuples of Expression
+        tuples (OrderedSet[TupleIDP]): OrderedSet of tuples of Expression
 
         constructors (List[Constructor]): List of Constructor
 
@@ -1029,7 +1030,7 @@ class ConstructedFrom(Enumeration):
                 (None, filter))
 
 
-class Tuple(ASTNode):
+class TupleIDP(ASTNode):
     def __init__(self, **kwargs):
         self.args = kwargs.pop('args')
         self.code = intern(",".join([str(a) for a in self.args]))
@@ -1041,7 +1042,7 @@ class Tuple(ASTNode):
         return self.code
 
 
-class FunctionTuple(Tuple):
+class FunctionTuple(TupleIDP):
     def __init__(self, **kwargs):
         self.args = kwargs.pop('args')
         if not isinstance(self.args, list):
@@ -1051,7 +1052,7 @@ class FunctionTuple(Tuple):
         self.code = intern(",".join([str(a) for a in self.args]))
 
 
-class CSVTuple(Tuple):
+class CSVTuple(TupleIDP):
     pass
 
 
@@ -1067,26 +1068,26 @@ class Ranges(Enumeration):
                 if x.fromI.type != self.type:
                     if self.type in [INT, REAL] and x.fromI.type in [INT, REAL]:
                         self.type = REAL  # convert to REAL
-                        tuples = [Tuple(args=[n.args[0].real()])
+                        tuples = [TupleIDP(args=[n.args[0].real()])
                                   for n in tuples]
                     else:
                         self.check(False,
                             f"incorrect value {x.fromI} for {self.type}")
 
                 if x.toI is None:
-                    tuples.append(Tuple(args=[x.fromI]))
+                    tuples.append(TupleIDP(args=[x.fromI]))
                 elif self.type == INT and x.fromI.type == INT and x.toI.type == INT:
                     for i in range(x.fromI.py_value, x.toI.py_value + 1):
-                        tuples.append(Tuple(args=[Number(number=str(i))]))
+                        tuples.append(TupleIDP(args=[Number(number=str(i))]))
                 elif self.type == REAL and x.fromI.type == INT and x.toI.type == INT:
                     for i in range(x.fromI.py_value, x.toI.py_value + 1):
-                        tuples.append(Tuple(args=[Number(number=str(float(i)))]))
+                        tuples.append(TupleIDP(args=[Number(number=str(float(i)))]))
                 elif self.type == REAL:
                     self.check(False, f"Can't have a range over real: {x.fromI}..{x.toI}")
                 elif self.type == DATE and x.fromI.type == DATE and x.toI.type == DATE:
                     for i in range(x.fromI.py_value, x.toI.py_value + 1):
                         d = Date(iso=f"#{date.fromordinal(i).isoformat()}")
-                        tuples.append(Tuple(args=[d]))
+                        tuples.append(TupleIDP(args=[d]))
                 else:
                     self.check(False, f"Incorrect value {x.toI} for {self.type}")
         Enumeration.__init__(self, tuples=tuples)
@@ -1343,7 +1344,7 @@ idpparser = metamodel_from_file(dslFile, memoization=True,
 
                                          Structure, SymbolInterpretation,
                                          Enumeration, FunctionEnum, CSVEnumeration,
-                                         Tuple, FunctionTuple, CSVTuple,
+                                         TupleIDP, FunctionTuple, CSVTuple,
                                          ConstructedFrom, Constructor, Ranges,
                                          Display,
 
