@@ -183,7 +183,7 @@ def _set_consequences_get_changed_choices(self):
     # clear consequences, as these might not be cleared by add_given when
     # running via CLI
     for a in self.assignments.values():
-        if a.status in [S.CONSEQUENCE]:
+        if a.status in [S.CONSEQUENCE, S.ENV_CONSQ]:
             self.assignments.assert__(a.sentence, None, S.UNKNOWN)
 
     removed_choices = {a.sentence.code: a for a in self.previous_assignments.values()
@@ -385,7 +385,7 @@ def _propagate_ignored(self, tag=S.CONSEQUENCE, given_todo=None):
     solver = self.solver_reified
     solver.push()
 
-    todo = (given_todo.copy() if given_todo else
+    todo = (deepcopy(given_todo) if given_todo else
             {a.sentence.code: a.sentence
             for a in self.assignments.values()
             if a.status not in [S.GIVEN, S.DEFAULT, S.EXPANDED] and
