@@ -23,17 +23,17 @@ Class to represent a collection of theory and structure blocks.
 """
 
 import time
-from copy import copy
+from copy import copy, deepcopy
 from enum import Enum, auto
 from itertools import chain
 from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union
 from z3 import (Context, BoolRef, ExprRef, Solver, sat, unsat, Optimize, Not, And, Or, Implies,
-                is_and, BoolVal)
+                is_and, BoolVal, get_param)
 
 from .Assignments import Status as S, Assignment, Assignments
 from .Expression import (TRUE, Expression, FALSE, AppliedSymbol,
                          EQUALS, NOT, Extension)
-from .Parse import (TypeDeclaration, Declaration, SymbolDeclaration, Symbol,
+from .Parse import (TypeDeclaration, Declaration, SymbolDeclaration, Symbol, SYMBOL,
                     TheoryBlock, Structure, Definition, str_to_IDP, SymbolInterpretation)
 from .Simplify import join_set_conditions
 from .utils import (OrderedSet, NEWL, BOOL, INT, REAL, DATE,
@@ -224,7 +224,7 @@ class Theory(object):
         """
         out = copy(self)
         out.assignments = self.assignments.copy()
-        out.constraints = OrderedSet(c.copy() for c in self.constraints)
+        out.constraints = OrderedSet(deepcopy(c) for c in self.constraints)
         out.def_constraints = {k:[e for e in v]  #TODO e.copy()
                                for k,v in self.def_constraints.items()}
         # copy() is called before making substitutions => invalidate derived fields
@@ -264,9 +264,9 @@ class Theory(object):
             if isinstance(block, TheoryBlock) or isinstance(block, Theory):
                 self.co_constraints = None
                 self.definitions += block.definitions
-                self.constraints.extend(v.copy() for v in block.constraints)
+                self.constraints.extend(deepcopy(v) for v in block.constraints)
                 self.def_constraints.update(
-                    {k:v.copy() for k,v in block.def_constraints.items()})
+                    {k:deepcopy(v) for k,v in block.def_constraints.items()})
 
         # apply the enumerations and definitions
 
@@ -291,7 +291,7 @@ class Theory(object):
                 symbol = t.args[0]
                 decl = self.declarations[symbol.name[1:]]
                 assert decl.instances, f"goal {decl.name} must be instantiable."
-                relevant = Symbol(name=RELEVANT)
+                relevant = SYMBOL(RELEVANT)
                 relevant.decl = self.declarations[RELEVANT]
                 for i in decl.instances.values():
                     constraint = AppliedSymbol.make(relevant, [i])
@@ -320,9 +320,6 @@ class Theory(object):
             if s.code not in self.assignments:
                 self.assignments.assert__(s, None, S.UNKNOWN)
 
-        for ass in self.assignments.values():
-            ass.sentence.original = ass.sentence
-
         self._constraintz = None
         return self
 
@@ -575,6 +572,8 @@ class Theory(object):
 
         - ``No models.``
 
+        - ``No model found in xx seconds.  Models may be available.  Change the timeout_seconds argument to see them.``
+
         - ``More models may be available.  Change the max argument to see them.``
 
         - ``More models may be available.  Change the timeout_seconds argument to see them.``
@@ -605,6 +604,8 @@ class Theory(object):
             solver = self.solver
             solver.push()
             self._add_assignment(solver)
+        default_timeout = get_param("timeout")
+
         todo = OrderedSet(a.sentence for a in self.get_core_atoms([S.UNKNOWN]))
 
         for q in todo:
@@ -614,13 +615,18 @@ class Theory(object):
         count, ass = 0, {}
         start = time.process_time()
         while ((max <= 0 or count < max) and
-               (timeout_seconds <= 0 or time.process_time()-start < timeout_seconds)):
+               (timeout_seconds <= 0 or time.process_time() - start < timeout_seconds)):
+            if timeout_seconds:
+                remaining = timeout_seconds - (time.process_time() - start)
+                solver.set("timeout", int(remaining*1000+200))
             # exclude ass
             different = []
             for a in ass.values():
                 if a.status == S.EXPANDED:
                     q = a.sentence
                     different.append(q.translate(self) != a.value.translate(self))
+            if 0 < count and len(different) == 0:
+                break
             if different:
                 solver.add(Or(different))
 
@@ -628,8 +634,6 @@ class Theory(object):
                 count += 1
                 ass = self._from_model(solver, todo, complete)
                 yield ass
-                if not todo:  # empty theory
-                    break
             else:
                 break
 
@@ -642,11 +646,15 @@ class Theory(object):
             param = ("max and timeout_seconds arguments" if maxed and timeouted else
                      "max argument" if maxed else
                      "timeout_seconds argument")
-            yield f"{NEWL}More models may be available.  Change the {param} to see them."
+            if count == 0:
+                yield f"{NEWL}No model found in {timeout_seconds} seconds.  Models may be available.  Change the {param} to see them."
+            else:
+                yield f"{NEWL}More models may be available.  Change the {param} to see them."
         elif 0 < count:
             yield f"{NEWL}No more models."
         else:
             yield "No models."
+        solver.set("timeout", int(default_timeout))
 
     def optimize(self,
                  term: str,
@@ -764,26 +772,39 @@ class Theory(object):
         range = termE.decl.range
         assert range, f"Can't determine range on infinite domains"
 
+        # consider every value in range
+        atoms = [Assignment(termE, val, S.UNKNOWN).formula() for val in range]
+        todos = {a.code: a for a in atoms}
+
+        # initialize the forbidden values
+        forbidden = set(str(e.sub_exprs[1]) for e in todos.values()
+                        if str(e) in self.assignments
+                        and self.assignments[str(e)].status in [S.GIVEN]
+                        and self.assignments[str(e)].value.same_as(FALSE))
+
         #  remove current assignments to same term
+        backup = self.assignments
+        self.assignments = self.assignments.copy()
+        removed = []
         if self.assignments[term].value:
             for k,a in self.assignments.items():
                 if (a.sentence.is_assignment and
                         a.sentence.code.startswith(term) and
                         a.status in [S.GIVEN, S.DEFAULT, S.EXPANDED]):
                     self.assert_(k, None, S.UNKNOWN)
+                    removed.append(a)
 
-        # consider every value in range
-        atoms = [Assignment(termE, val, S.UNKNOWN).formula() for val in range]
-        todos = {a.code: a for a in atoms}
-
-        forbidden = set()
         for ass in self._propagate(given_todo=todos):
             if isinstance(ass, str):
                 continue
             if ass.value.same_as(FALSE):
                 forbidden.add(str(ass.sentence.sub_exprs[1]))
 
-        return [str(e.sub_exprs[1]) for e in todos.values() if str(e.sub_exprs[1]) not in forbidden]
+        # restore the assignments
+        self.assignments = backup
+
+        return [str(e.sub_exprs[1]) for e in todos.values()
+                if str(e.sub_exprs[1]) not in forbidden]
 
     def explain(self,
                 consequence: Optional[str] = None
@@ -922,7 +943,7 @@ class Theory(object):
                        max_rows: int = 50,
                        first_hit: bool = True,
                        verify: bool = False
-                       ) -> (List[List[Assignment]], bool):
+                       ) -> Tuple[List[List[Assignment]], bool]:
         """Experimental.  Returns the rows for a decision table that defines ``goal_string``.
 
         ``goal_string`` must be a predicate application defined in the theory.
